import abc

class ColorPallete(metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def main_color():
        pass

    @abc.abstractproperty
    def secondry():
        pass

    @abc.abstractproperty
    def reg():
        pass
    
    @abc.abstractproperty
    def drop():
        pass

class Default(ColorPallete):
    main_color = "#c0cacc"
    secondry   = "#a6b1b3"
    reg        = "#7ba1a6"
    drop       = "#7ba1a6"

class Blues(ColorPallete):
    main_color = "#4F7C91"
    secondry   = "#6F9AAB"
    reg        = "#2E6583"
    drop       = "#4F7C91"

class Grays(ColorPallete):
    main_color = "#9ea3a3"
    secondry   = "#909696"
    reg        = "#828989"
    drop       = "#747c7c"

class Reds(ColorPallete):
    main_color = "#ca1414"
    secondry   = "#e9a8a1"
    reg        = "#f4d4d3"
    drop       = "#6b0003"

class Greens(ColorPallete):
    main_color = "#5c715e"
    secondry   = "#87a08b"
    reg        = "#d2d6a8"
    drop       = "#3a5335"


import math
from PIL import ImageDraw

class Connector:
    def __init__(self) -> None:
        pass

    def connect(self, image, layer1, layer2):
        _from = layer1.get_from()
        _to   = layer2.get_to()

        draw = ImageDraw.Draw(image)
        # [(x, y), (x, y)]
        for p1, p2 in zip(_from, _to):
            draw.line([p1, p2], fill="#000000", width=5)

        return image
    
    def output(self, image, layer1):
        _from = layer1.get_from()
        draw = ImageDraw.Draw(image)

        points1 = [(_from[0][0]      , _from[0][1] + 350),
                   (_from[0][0] + 200, _from[0][1] + 350),
                   (_from[0][0] + 150, _from[0][1] + 300),
                   (_from[0][0] + 200, _from[0][1] + 350),
                   (_from[0][0] + 150, _from[0][1] + 400)]
        
        points2 = [(_from[1][0]     , _from[1][1] - 350),
                   (_from[1][0] + 200, _from[1][1] - 350),
                   (_from[1][0] + 150, _from[1][1] - 300),
                   (_from[1][0] + 200, _from[1][1] - 350),
                   (_from[1][0] + 150, _from[1][1] - 400)]
        draw.line(points1, fill="#000000", width=5)
        draw.line(points2, fill="#000000", width=5)


from keras.layers import Layer
from litten.layers import *
from litten.visualize.palettes import *


palettes = {
    "default":Default,
    "red"    :Reds,
    "blue"   :Blues,
    "gray"   :Grays,
    "green"  :Greens
}

activations = ["Activation", "Softmax", "ELU", "ReLU", "PReLU", "LeakyReLU", "ThresholdedReLU"]
convs       = ["Conv1D", "Conv2D", "Conv3D", "SeparableConv1D", "SeparableConv2D", "SeparableConv3D", "DepthwiseConv2D", "Conv1DTranspose", "Conv2DTranspose", "Conv3DTranspose" ]
pools       = ["MaxPooling1D", "MaxPooling2D", "MaxPooling3D", "AveragePooling1D", "AveragePooling2D", "AveragePooling3D", "GlobalMaxPooling1D", "GlobalMaxPooling2D", "GlobalMaxPooling3D", "GlobalAveragePooling1D", "GlobalAveragePooling2D", "GlobalAveragePooling3D"]
rnns        = ["LSTM", "GRU", "RNN", "SimpleRNN", "Bidirectional"]
convlstms   = ["ConvLSTM1D layer", "ConvLSTM2D layer", "ConvLSTM3D layer"]

def get_width(layers):
    width = 0
    layer_name = get_layer_name(layer=layers[0])
    if layer_name == "Input": 
        del layers[0]

    width += 60

    for layer in layers:        
        layer_name = get_layer_name(layer=layer)
        if layer_name in activations:
            width += 70
        elif layer_name in convs or layer_name in pools or layer_name in convlstms:
            width += 180
        elif layer_name in rnns or layer_name == 'Flatten':
            width += 140
        else:
            width += 80
    return width + 60   

def get_layer_name(layer):
    try:
        return layer.__class__.__name__
    except:
        return "unknown"

def get_activation_name(layer):
    try:
        return layer.activation.__name__
    except:
        return "unknown"


import math
from matplotlib import font_manager
from PIL import ImageDraw, ImageFont 


fonts = font_manager.findSystemFonts(fontpaths=None, fontext='ttf')
font = ImageFont.truetype(font=fonts[0], size=100)


class Layer:
    """
    Layer class contain the base structure for any layer
    """
    def __init__(self, name: str, start_x: int, palette) -> None:
        """
        Construct Layer class

        Args:
            name: layer name
            number: layer number
        """
        self.name     = name
        self._start_x = start_x
        self._end_x   = 0
        self.palette  = palette

    def draw(self, image, show_name = False, show_properties=False):
        """
        draw default layer 
        """
        draw = ImageDraw.Draw(image)
        points = [(self._start_x + 2000, 800),
                  (self._start_x + 60, 200)]
        
        draw.rounded_rectangle(xy  = points, radius=3, fill=self.palette.main_color)
        
        draw.ellipse((points[0][0] + 100, points[0][1] + 150, points[0][0] + 300, points[0][1] + 350 ), fill = '#ffffff', outline='#000')
        draw.ellipse((points[0][0] + 100, points[0][1] + 500, points[0][0] + 300, points[0][1] + 700 ), fill = '#ffffff', outline='#000')
        draw.ellipse((points[0][0] + 100, points[0][1] + 850, points[0][0] + 300, points[0][1] + 1050), fill = '#ffffff', outline='#000')

        # update end_x
        self._end_x = self._start_x + 800

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

    @property
    def end(self):
        return self._end_x

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = (self._start_x + 200, 2100)
        draw.text(points, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)
        points = (self._start_x + 200, 2100)
        draw.text(points, text=self.name, fill="#000000", font=font)

    def get_from(self):
        return [(self._start_x + 600, 800), (self._start_x + 600, 2000)]

    def get_to(self):
        return [(self._start_x + 200, 800), (self._start_x + 200, 2000)]


class InputLayer(Layer):
    """
    Input Layer
    """
    def __init__(self, name, shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.shape = shape
    
    def draw(self, image, show_name=False, show_properties = False):
        """
        Draw input layer on the image

        Args:
            image: The image will draw the layer on it
        
        Returns:
            image: The image after drawing the layer
        """
        draw = ImageDraw.Draw(image)

        for i in range(10):
            points = [(self._start_x + 200, 400 + i * 200), (self._start_x + 400, 600 + i * 200)] 
            draw.rectangle(points, fill = self.palette.main_color, width = 5, outline= '#000000')

        # update end_x
        self._end_x = self._start_x + 600

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = (self._start_x + 150, 2500)
        draw.text(points, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)

        points = (self._start_x + 150, 2500)        
        draw.text(points, text=self.name, fill="#000000", font=font) 
        
        points = (self._start_x, 2620)
        draw.text(points, text="{}".format(self.shape), fill="#000000", font=font) 

    def get_from(self):
        return [(self._start_x + 400, 400), (self._start_x + 400, 2400)]
        

class DenseLayer(Layer):
    """
    Fully Connected Layer (Dense Layer)
    """
    def __init__(self, name, units, activation, input_shape, output_shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.units = units
        self.activation   = activation
        self.input_shape  = input_shape
        self.output_shape = output_shape

    def draw(self, image, show_name = False, show_properties=False):
        """
        Draw a dense layer (FullyConnectedLayer)
        Args:
            image: The image will draw the layer on it
        
        Returns:
            image: The image after drawing the layer
        """
        draw = ImageDraw.Draw(image)
        points = [(self._start_x + 200, 600),
                  (self._start_x + 600, 2200)]
        
        draw.rounded_rectangle(xy  = points, radius=5, fill=self.palette.main_color)
        
        draw.ellipse((points[0][0] + 100, points[0][1] + 150 , points[0][0] + 300, points[0][1] + 350 ), fill = '#ffffff', outline='#000')
        draw.ellipse((points[0][0] + 100, points[0][1] + 500 , points[0][0] + 300, points[0][1] + 700 ), fill = '#ffffff', outline='#000')
        draw.ellipse((points[0][0] + 100, points[0][1] + 1200, points[0][0] + 300, points[0][1] + 1400), fill = '#ffffff', outline='#000')

        draw.line   ([(points[0][0] + 200, points[0][1] + 800 ), (points[0][0] + 200, points[0][1] + 810 )], fill='#000000', width=20)
        draw.line   ([(points[0][0] + 200, points[0][1] + 900 ), (points[0][0] + 200, points[0][1] + 910 )], fill='#000000', width=20)
        draw.line   ([(points[0][0] + 200, points[0][1] + 1000), (points[0][0] + 200, points[0][1] + 1010)], fill='#000000', width=20)
        draw.line   ([(points[0][0] + 200, points[0][1] + 1100), (points[0][0] + 200, points[0][1] + 1110)], fill='#000000', width=20)

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = self._start_x + 800

        return image
    
    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points1 = (self._start_x + 230, 2250)
        draw.text(points1, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)
        points1 = (self._start_x + 230, 2250)
        draw.text(points1, text=self.name, fill="#000000", font=font)
        
        points2 = (self._start_x + 230, 2370)
        draw.text(points2, text="units: {}".format(self.units), fill="#000000", font=font)
        
        points3 = (self._start_x + 230, 2490)
        draw.text(points3, text=self.activation, fill="#000000", font=font)

        points4 = (self._start_x + 230, 2610)
        draw.text(points4, text="input shape\n{}".format(self.input_shape), fill="#000000", font=font)        
        
        points5 = (self._start_x + 230, 2850)
        draw.text(points5, text="output shape\n{}".format(self.output_shape), fill="#000000", font=font)

    def get_from(self):
        return [(self._start_x + 600, 630 ),
                (self._start_x + 600, 2170)]
    
    def get_to(self):
        return [(self._start_x + 200, 630 ),
                (self._start_x + 200, 2170)]


class ConvLayer(Layer):
    """
    Convluation Layer 
    """
    def __init__(self, name, filters, kernel, activation, input_shape, output_shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.filters = filters
        self.kernel  = kernel
        self.activation   = activation
        self.input_shape  = input_shape
        self.output_shape = output_shape
        self._c     = int(min(math.log(filters, 2), 10))
        self._s     = min(int(input_shape[1]/100), 10)
        self.lpoints= [] 

    def draw(self, image, show_name = False, show_properties=False):
        draw = ImageDraw.Draw(image)
        points = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                  self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]

        for i in range(self._c):
            if self._c % 2 == 0:
                if i%2:
                    draw.rectangle(points, fill = self.palette.main_color, outline='#000000')
                else:
                    draw.rectangle(points, fill = self.palette.secondry, outline='#000000')
            else:
                if i%2:
                    draw.rectangle(points, fill = self.palette.secondry, outline='#000000')
                else:
                    draw.rectangle(points, fill = self.palette.main_color, outline='#000000')

            points[0], points[1], points[2], points[3] = points[0] + 50, points[1] + 50, points[2] + 50, points[3] + 50
        
        self.lpoints = points

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = points[2] + 200

        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = [self.lpoints[0], self.lpoints[3]]
        draw.text(points, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)

        points = [self.lpoints[0], self.lpoints[3]]
        draw.text(points, text=self.name, fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 120]
        draw.text(points, text=self.activation, fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 240]
        draw.text(points, text="filters:{}".format(self.filters), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 360]
        draw.text(points, text="kernel:{}".format(self.kernel), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 480]
        draw.text(points, text="input shape:\n{}".format(self.input_shape), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 720]
        draw.text(points, text="output shape:\n{}".format(self.output_shape), fill="#000000", font=font)

    def get_from(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        
        point_end   = [self._start_x + 300                         + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s        + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]

        return [(point_end[2], point_end[1]),
                (point_end[2], point_end[3]),
                (point_start[2], point_start[1])]
    
    def get_to(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        point_end   = [self._start_x + 300                      + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s     + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]


        return [(point_end[0], point_end[1]),
                (point_end[0], point_end[3]),
                (point_start[0], point_start[1])]


class PoolingLayer(Layer):
    """
    Pooling Layer
    """
    def __init__(self, name, pool_size, padding, input_shape, output_shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.units = 64
        self.pool_size = pool_size
        self.padding   = padding
        self.input_shape  = input_shape
        self.output_shape = output_shape
        self._c      = int(min(math.log(self.units, 2), 10))
        self._s      = int(input_shape[1]/100) 
        self.lpoints = []

    def draw(self, image, show_name=False, show_properties=False):
        draw = ImageDraw.Draw(image)
        points = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                  self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]

        for i in range(self._c):
            if self._c % 2 == 1:
                if i%2:
                    draw.rectangle(points, fill = self.palette.main_color, outline='#000000')
                else:
                    draw.rectangle(points, fill = self.palette.secondry, outline='#000000')
            else:
                if i%2:
                    draw.rectangle(points, fill = self.palette.secondry, outline='#000000')
                else:
                    draw.rectangle(points, fill = self.palette.main_color, outline='#000000')

            points[0], points[1], points[2], points[3] = points[0] + 50, points[1] + 50, points[2] + 50, points[3] + 50
        
        self.lpoints = points
        
        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = points[2] + 200

        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = [self.lpoints[0], self.lpoints[3]]
        draw.text(points, text=self.name, fill="#000000", font=font)
    
    def _show_prop(self, image):

        draw = ImageDraw.Draw(image)
        
        points = [self.lpoints[0], self.lpoints[3]]
        draw.text(points, text=self.name, fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 120]
        draw.text(points, text="padding: {}".format(self.padding), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 240]
        draw.text(points, text="pool: {}".format(self.pool_size), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 360]
        draw.text(points, text="input shape: \n{}".format(self.input_shape), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 600]
        draw.text(points, text="output shape \n{}".format(self.output_shape), fill="#000000", font=font)

    def get_from(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        
        point_end   = [self._start_x + 300                       + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s      + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]

        return [(point_end[2], point_end[1]),
                (point_end[2], point_end[3]),
                (point_start[2], point_start[1])]
    
    def get_to(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        point_end   = [self._start_x + 300                        + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s       + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]


        return [(point_end[0], point_end[1]),
                (point_end[0], point_end[3]),
                (point_start[0], point_start[1])]


class EmbeddingLayer(Layer):
    """
    Embedding Layer
    """
    def __init__(self, name, input_dim, output_dim, input_shape, output_shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.input_dim    = input_dim
        self.output_dim   = output_dim
        self.input_shape  = input_shape
        self.output_shape = output_shape
        self.lpoints      = []

    def draw(self, image, show_name=False, show_properties=False):
        draw = ImageDraw.Draw(image)

        for i in range(7):
            points = [(self._start_x + 200, 700 + i * 200), (self._start_x + 400, 900 + i * 200)] 
            if i % 2:
                draw.rounded_rectangle(points, radius = 2, fill = self.palette.main_color, width = 5, outline= '#000000')
            else:
                draw.rounded_rectangle(points, radius = 2, fill = self.palette.secondry  , width = 5, outline= '#000000')

        for i in range(7):
            points = [(self._start_x + 400, 700 + i * 200), (self._start_x + 600, 900 + i * 200)]
            if i % 2:
                draw.rounded_rectangle(points, radius = 2, fill = self.palette.secondry,   width = 5, outline= '#000000')
            else:
                draw.rounded_rectangle(points, radius = 2, fill = self.palette.main_color, width = 5, outline= '#000000')

        self.lpoints = points

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = self._start_x + 800
        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = [self.lpoints[0][0] - 2500, self.lpoints[1][1] + 500]
        draw.text(points, text=self.name, fill="#000000", font=font)
    
    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)
        
        points = [self.lpoints[0][0] - 270, self.lpoints[1][1] + 50]
        draw.text(points, text=self.name, fill="#000000", font=font)

        points = [self.lpoints[0][0] - 300, self.lpoints[1][1] + 170]
        draw.text(points, text="input_dim :{}".format(self.input_dim), fill="#000000", font=font)

        points = [self.lpoints[0][0] - 300, self.lpoints[1][1] + 290]
        draw.text(points, text="output_dim:{}".format(self.output_dim), fill="#000000", font=font)

        points = [self.lpoints[0][0] - 300, self.lpoints[1][1] + 410]
        draw.text(points, text="input_shape:\n{}".format(self.input_shape), fill="#000000", font=font)

        points = [self.lpoints[0][0] - 300, self.lpoints[1][1] + 650]
        draw.text(points, text="output_shape:\n{}".format(self.output_shape), fill="#000000", font=font)
 
    def get_from(self):
        return [(self._start_x + 570, 700), (self._start_x + 570, 2100)]
    
    def get_to(self):
        return [(self._start_x + 230, 700), (self._start_x + 230, 2100)]


class RecurrentLayer(Layer):
    """
    Recurrent Layer
    """
    def __init__(self, name, units, activation, bi, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.units = units
        self.activation = activation
        self.bi = bi
        self.lpoints = []

    def draw(self, image, show_name=False, show_properties=False):
        draw = ImageDraw.Draw(image)
        
        points1 = [ self._start_x + 400, 1200, self._start_x + 1200, 1800]
        points2 = [(self._start_x + 400, 1200), (self._start_x + 300, 1100), (self._start_x + 300, 1700), (self._start_x + 400, 1800)]
        points3 = [(self._start_x + 400, 1200), (self._start_x + 300, 1100), (self._start_x + 1100, 1100), (self._start_x + 1200, 1200)]

        draw.rectangle(points1, fill = self.palette.main_color, outline='#000000')
        draw.polygon  (points2, fill = self.palette.main_color, outline='#000000')
        draw.polygon  (points3, fill = self.palette.main_color, outline='#000000')

        # draw arrows
        points5 = [(self._start_x + 600 , 1400), (self._start_x + 1000, 1400), (self._start_x + 950, 1350), (self._start_x + 1000, 1400), (self._start_x + 950, 1450)]
        points6 = [(self._start_x + 1000, 1600), (self._start_x + 600 , 1600), (self._start_x + 650, 1550), (self._start_x + 600 , 1600), (self._start_x + 650, 1650)]

        points7 = [(self._start_x + 600 , 1500), (self._start_x + 1000, 1500), (self._start_x + 950, 1450), (self._start_x + 1000, 1500), (self._start_x + 950, 1550)]

        self.lpoints = points1
        if self.bi:
            draw.line(points5, fill="#000000")
            draw.line(points6, fill="#000000")
        else:
            draw.line(points7, fill="#000000")
        
        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = self._start_x + 1400
        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        draw.text((self.lpoints[0], self.lpoints[3] + 50) , text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)

        draw.text((self.lpoints[0], self.lpoints[3] + 50 ), text=self.name, fill="#000000", font=font)
        draw.text((self.lpoints[0], self.lpoints[3] + 170), text="units: {}".format(self.units), fill="#000000", font=font)
        draw.text((self.lpoints[0], self.lpoints[3] + 290), text=self.activation, fill="#000000", font=font)

    def get_from(self):
        return [(self._start_x + 1200, 1200),
                (self._start_x + 1200, 1800),
                (self._start_x + 1100, 1100)]
    
    def get_to(self):
        return [(self._start_x + 400, 1200),
                (self._start_x + 400, 1800),
                (self._start_x + 300, 1100)]


class ConvLSTM(Layer):
    """
    ConvLSTM Layer
    """
    def __init__(self, name, filters, kernels, activation, shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.filters    = filters
        self.kernels    = kernels
        self.activation = activation
        self.shape = shape
        self._c    = int(min(math.log(filters, 2), 10))
        self._s    = min(int(shape[0]/100), 10)
        self.lpoints = []

    def draw(self, image, show_name=False, show_properties=False):
        draw = ImageDraw.Draw(image)
        points = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                  self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]

        for i in range(self._c):
            if i%2:
                draw.rectangle(points, fill = self.palette.main_color, outline='#000000')
            else:
                draw.rectangle(points, fill = self.palette.secondry, outline='#000000')
            points[0], points[1], points[2], points[3] = points[0] + 50, points[1] + 50, points[2] + 50, points[3] + 50
        
        points[0], points[1], points[2], points[3] = points[0] - 50, points[1] - 50, points[2] - 50, points[3] - 50


        points2 = [(points[0] +      (points[2] - points[0]) // 5,      points[1] + (points[3] - points[1]) // 2),
                   (points[0] + 40 * (points[2] - points[0]) // 5,      points[1] + (points[3] - points[1]) // 2),
                   (points[0] + 40 * (points[2] - points[0]) // 5 - 50, points[1] + (points[3] - points[1]) // 2 - 50),
                   (points[0] + 40 * (points[2] - points[0]) // 5     , points[1] + (points[3] - points[1]) // 2     ),
                   (points[0] + 40 * (points[2] - points[0]) // 5 - 50, points[1] + (points[3] - points[1]) // 2 + 50)]

        self.lpoints = points
        draw.line(points2, fill="#000000")
        
        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = points[2] + 200

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = [self.lpoints[0], self.lpoints[3] + 50]
        draw.text(points, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)

        points = [self.lpoints[0], self.lpoints[3] + 50]
        draw.text(points, text=self.name, fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 170]
        draw.text(points, text="filters:{}".format(self.filters), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 290]
        draw.text(points, text="kernel :{}".format(self.kernels), fill="#000000", font=font)

        points = [self.lpoints[0], self.lpoints[3] + 410]
        draw.text(points, text="{}".format(self.activation), fill="#000000", font=font)
    
    def get_from(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        
        point_end   = [self._start_x + 300                       + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s      + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]

        return [(point_end[2], point_end[1]),
                (point_end[2], point_end[3]),
                (point_start[2], point_start[1])]

    def get_to(self):
        point_start = [self._start_x + 300                 , 1100 - 50 * self._c // 2 - 100 * self._s // 2,
                       self._start_x + 1000 + 100 * self._s, 1700 - 50 * self._c // 2 + 100 * self._s // 2]
        point_end   = [self._start_x + 300                        + 50 * (self._c - 1), 1100 - 50 * self._c // 2 - 100 * self._s // 2 + 50 * (self._c - 1),
                       self._start_x + 1000 + 100 * self._s       + 50 * (self._c - 1), 1700 - 50 * self._c // 2 + 100 * self._s // 2 + 50 * (self._c - 1)]


        return [(point_end[0], point_end[1]),
                (point_end[0], point_end[3]),
                (point_start[0], point_start[1])]
    

class ActivationLayer(Layer):
    """
    Activation Layer
    """
    def __init__(self, name, activation, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.activation = activation

    def draw(self, image, show_name=False, show_properties=False):
        draw = ImageDraw.Draw(image)

        draw.rectangle((self._start_x + 200, 1600, self._start_x + 600, 1200), fill = self.palette.main_color)
        draw.ellipse  ((self._start_x + 250, 1250, self._start_x + 550, 1550), fill = '#ffffff', outline='#000000')

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)

        # update end_x
        self._end_x = self._start_x + 700
        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = (self._start_x + 250, 1650)
        draw.text(points, text=self.name, fill="#000000", font=font)

    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)
        points = (self._start_x + 250, 1650)
        draw.text(points, text=self.name, fill="#000000", font=font)
        if self.activation:
            points = (self._start_x + 250, 1770)
            draw.text(points, text=self.activation, fill="#000000", font=font)

    def get_from(self):
        return [(self._start_x + 600, 1200), (self._start_x + 600, 1600)]
    
    def get_to(self):
        return [(self._start_x + 200, 1200), (self._start_x + 200, 1600)]


class FlattenLayer(Layer):
    """
    Flatten Layer
    """
    def __init__(self, name, input_shape, output_shape, start_x, palette) -> None:
        super().__init__(name, start_x, palette)
        self.input_shape = input_shape
        self.output_shape = output_shape

    def draw(self, image, show_name=False, show_properties=False):

        draw = ImageDraw.Draw(image)

        points1 = [ self._start_x + 100 , 1400 ,  self._start_x + 1100, 1600]
        points2 = [(self._start_x + 100 , 1400), (self._start_x + 200 , 1300), (self._start_x + 1200, 1300), (self._start_x + 1100, 1400)]
        points3 = [(self._start_x + 1200, 1300), (self._start_x + 1100, 1400), (self._start_x + 1100, 1600), (self._start_x + 1200, 1500)]

        draw.rectangle(points1, fill = self.palette.main_color, width = 3, outline='#000000')
        draw.polygon  (points2, fill = self.palette.main_color, width = 3, outline='#000000')
        draw.polygon  (points3, fill = self.palette.main_color, width = 3, outline='#000000')

        if show_properties:
            self._show_prop(image)
        elif show_name:
            self._show_name(image)
            
        # update end_x
        self._end_x = self._start_x + 1400
        return image

    def _show_name(self, image):
        draw = ImageDraw.Draw(image)
        points = [self._start_x + 100, 1650]
        draw.text(points, text=self.name, fill="#000000", font=font)
    
    def _show_prop(self, image):
        draw = ImageDraw.Draw(image)

        points = [self._start_x + 200, 1650]
        draw.text(points, text=self.name, fill="#000000", font=font)

        points = [self._start_x + 200, 1770]
        draw.text(points, text="input shape:\n{}".format(self.input_shape), fill="#000000", font=font)


        points = [self._start_x + 200, 2010]
        draw.text(points, text="output shape:\n{}".format(self.output_shape), fill="#000000", font=font)

    def get_from(self):
        return [(self._start_x + 1100, 1400),
                (self._start_x + 1100, 1600),
                (self._start_x + 1200, 1300)]
    
    def get_to(self):
        return [(self._start_x + 100, 1400),
                (self._start_x + 100, 1600),
                (self._start_x + 200, 1300)]


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.7'
DESCRIPTION = 'A python package to visualize sequentail tensorflow model archtictures'

# Setting up
setup(
    name="litten",
    version=VERSION,
    author="Hossam Asaad",
    author_email="hossamasaad10@gmail.com",
    description=DESCRIPTION,
    long_description=long_description,  
    long_description_content_type="text/markdown",
    packages=find_packages(),
    keywords=['python', 'tensorflow', 'keras', 'PIL', 'pillow', 'visualize models', 'visualize'],

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'keras>=2.9.0',
        'ipython==8.7.0',
        'Pillow>=9.3.0',
        'numpy>=1.23.0',
        'matplotlib==3.5.2'
    ],
    python_requires='>=3.6',

)


import os
import sys
sys.path.append(os.path.realpath(''))

import numpy as np
import matplotlib.pyplot as plt
import PIL.Image as Image
from tensorflow.keras.models import Model

from litten import utils
from litten.visualize import palettes
from litten.layers import *
from IPython.display import display


class ModelVisualizer:
    def __init__(self, model, background_color = "#FFFFFF", palette = 'default', show_connectors=False, show_names=False, show_proporties=False) -> None:
        self.model            = model
        self.background_color = background_color
        self.show_names       = show_names
        self.show_proporties  = show_proporties
        self.show_connectors  = show_connectors
        self.width     = utils.get_width(model.layers) * 10
        self.height    = 3200
        self.palette   = utils.palettes[palette]
        self.image     = Image.new("RGB", (self.width, self.height), color=self.background_color)
        self.connector = Connector()

    def visualize_model(self):
        
        layers = self.model.layers
        last_layer = None

        # Draw input layer
        shape = None
        layer_name = utils.get_layer_name(layer=layers[0])

        if layer_name == "Input": 
            shape = layers[0].shape
            del layers[0]
        else:
            shape = layers[0].input_shape

        last_layer = InputLayer(name="Input", shape=shape, start_x=20, palette=self.palette)
        last_layer.draw(image=self.image, show_name=self.show_names, show_properties=self.show_proporties)
        curr_layer = None

        for layer in layers:
            
            layer_name = utils.get_layer_name(layer=layer)

            if layer_name == "Flatten":
                curr_layer = FlattenLayer(name=layer_name, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)
            
            elif layer_name == "Dense":
                activation = utils.get_activation_name(layer=layer)
                curr_layer = DenseLayer(name=layer_name, units=layer.units, activation=activation, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)
            
            elif layer_name == "Embedding":
                curr_layer = EmbeddingLayer(name=layer_name, input_dim=layer.input_dim, output_dim=layer.output_dim, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)

            elif layer_name in utils.activations:
                activation = None
                if layer_name == "Activation":
                    activation = utils.get_activation_name(layer=layer)
                curr_layer = ActivationLayer(name=layer_name, activation=activation, start_x=last_layer.end, palette=self.palette)

            elif layer_name in utils.convs:
                activation = utils.get_activation_name(layer=layer)
                curr_layer = ConvLayer(name=layer_name, filters=layer.filters, kernel=layer.kernel_size, activation=activation, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)

            elif layer_name in utils.pools:
                curr_layer = PoolingLayer(name=layer_name, pool_size=layer.pool_size, padding=layer.padding, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)

            elif layer_name in utils.rnns:
                bi = False
                if layer_name == 'Bidirectional':
                    bi = True
                    layer = layer.layer
                    layer_name = "Bi(" + utils.get_layer_name(layer) + ")"

                activation = utils.get_activation_name(layer=layer)
                curr_layer = RecurrentLayer(name=layer_name, units=layer.units, activation=activation, bi=bi, start_x=last_layer.end, palette=self.palette)

            elif layer_name in utils.convlstms:
                activation = utils.get_activation_name(layer=layer)
                curr_layer = ConvLayer(name=layer_name, filters=layer.filters, kernel=layer.kernel_size, activation=activation, input_shape=layer.input_shape, output_shape=layer.output_shape, start_x=last_layer.end, palette=self.palette)

            else:
                curr_layer = Layer(layer_name, start_x=last_layer.end, palette=self.palette)

            curr_layer.draw(image=self.image, show_name=self.show_names, show_properties=self.show_proporties)

            if self.show_connectors:
                self.connector.connect(image=self.image, layer1=last_layer, layer2=curr_layer)

            last_layer = curr_layer
        
        if self.show_connectors:
            self.connector.output(image=self.image, layer1=last_layer)

        scale = 1
        display(self.image.resize(( int(self.image.width * scale), int(self.image.height * scale))))


    def visualize_featuremap(self, input_image, cmap = "gray"):

        image = np.expand_dims(input_image, axis=0)
        
        fig = plt.figure()
        fig.suptitle("{}".format("Input Image") , fontsize=18)
        plt.imshow(input_image)

        for layer in self.model.layers:  

            if 'conv' not in layer.name:
                continue    
            
            
            intermediate_model = Model(inputs=self.model.inputs , outputs=layer.output)
            features = intermediate_model.predict(image)

            fig = plt.figure(figsize=(20, 20))
            fig.suptitle("{}".format(layer.name) , fontsize=18)

            for i in range(1, features.shape[3]+1):
                plt.subplot(8,8,i)
                plt.imshow(features[0,:,:,i-1] , cmap=cmap)

            plt.show()
    
    def visualize_filters(self, cmap = "gray"):
    
        for layer in self.model.layers:  

            if 'conv' not in layer.name:
                continue    
            
            filters, _ = layer.get_weights()

            fig = plt.figure(figsize=(10,10))
            fig.suptitle("{}".format(layer.name) , fontsize=18)

            idx=1
            no_filters = 6

            for i in range(no_filters):
                filter = filters[:,:,:,i]
                ch = min(filter.shape[2], 3)
                for j in range(ch):
                    plt.subplot(no_filters, ch, idx)
                    plt.imshow(filter[:,:,j], cmap=cmap)
                    idx += 1
            plt.show()



import os
import sys
sys.path.append(os.path.realpath(''))
from litten import utils 

class LayersSummary:
    def __init__(self) -> None:
        pass

    def show_layers_summaries(self, model):
        for i, layer in enumerate(model.layers):
            print("=================================================================================================================")
            x = "Layer {}: {}".format(i+1, utils.get_layer_name(layer))
            x = x + " " * (24 - len(x)) + "| Attributes"
            print(x)
            print("----------------------------------------")
            attributes = vars(layer)

            for key, value in attributes.items():
                s = key + " " * (23 - len(key))
                if key[0] != "_" and key != "kernel" and key != "bias":
                    print(s, ": ", value)